/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License") you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code in this file was released on September 25, 2004
 * 
 * Unless otherwise stated, the Initial Developer of the
 * Original Code is David Perry.  Portions created by David Perry are
 * Copyright (C) 2002 David Perry.  All rights reserved.
 * 
 * Contributor(s):
 *   David Perry <d.perry@utoronto.ca> (Original Author)
 *
 * Alternatively, the contents of this file may be used under the
 * terms of the GNU General Public License Version 2 or later (the
 * "GPL"), in which case the provisions of the GPL are applicable
 * instead of those above.  If you wish to allow use of your
 * version of this file only under the terms of the GPL and not to
 * allow others to use your version of this file under the MPL,
 * indicate your decision by deleting the provisions above and
 * replace them with the notice and other provisions required by
 * the GPL.  If you do not delete the provisions above, a recipient
 * may use your version of this file under either the MPL or the
 * GPL.
 */

/*
 * Global variables
 */

var gAuwrEditor;        // Editor object (?? see below)
var gAuWrapped;         // Flag - have we wrapped already?

/*
 * autoRewrapLoad() - window onload handler
 */
function autoRewrapLoad() {
    dump( "autoRewrapLoaded()\n" );

    var compStateListener = new auwrComposeStateListener();
    gMsgCompose.RegisterStateListener(compStateListener);
}

/*
 * autoRewrapReload() - window compose-window-reload handler
 */
function autoRewrapReload() {
    gAuWrapped = false;
}

/*
 * autoRewrap() - do the dirty deed
 */
function autoRewrap() {
    dump( "autoRewrap()\n" );
    gAuWrapped = true;
    try {
        if(window.GetCurrentEditor)
        {
            var doIt = true;
            var ed = window.GetCurrentEditor();
            var cbSaver = new auwrClipboardSaver();
            var prefs = new auwrPrefs();

            /* We want to re-select the current selection after we
                mess with the document.  First we need to save it */
            ed.beginTransaction();
            cbSaver.save();
            var selection = ed.selection;
            var ranges = [];
            var fudgeRanges = false;
            if( selection && selection.rangeCount )
            {
                for( var i=0; i < selection.rangeCount; i++ )
                {
                    var r = selection.getRangeAt(i);
                    ranges.push( { range:r, start: r.startOffset, end:r.endOffset, collapsed: r.collapsed } );
                }
            }

            /* Get the body text so we can play with it
               1026 = 2 (output formatted) | 1024 (LF linebreak)
               taken from enigmailMsgComposeOverlay.js */
            var docText = ed.outputToString("text/plain", 1026);

            var wrapGrace = prefs.getInt("graceMargin");
            if( wrapGrace )
            {
                var docLines = docText.split('\n');
                var longestLine = 0;
                for(var i in docLines)
                {
                    if(docLines[i].length > longestLine)
                        longestLine = docLines[i].length;
                }
                dump( "autoRewrap: longestLine="+longestLine+
                    ", wrapWidth="+ed.wrapWidth+
                    ", wrapGrace="+wrapGrace+"\n");

                if( longestLine <= ed.wrapWidth+wrapGrace )
                    doIt = false;
            }

            if( doIt )
            {
                // Avoid wrapping the sig if we can help it (bug 7379)
                var sigOfs = 0;
                if( !prefs.getBool("wrapSig") )
                {
                    // XXX: is this the best way to find the sig line?
                    sigOfs = docText.search(/[\r\n]-- +[\r\n]/);
                }
                if( sigOfs > 0 )
                {
                    sigOfs++;   // that extra newline before the dashes
                    var oldLength = docText.length;
                    var sigText = docText.substr(sigOfs);
                    docText = docText.substr(0,sigOfs);
                    ed.selectAll();
                    ed.insertTextWithQuotations(docText);
                    ed.rewrap(true);
                    ed.insertText(sigText);
                    var revisedText = ed.outputToString("text/plain", 1026);
                    // hack: rewrapped body can be bigger than it was, so
                    // it doesn't all get re-selected
                    fudgeRanges = revisedText.length-oldLength;

                }
                else
                {
                    // 'Dumb' rewrap
                    ed.rewrap(true);
                }
            }

            /* Restore the last selection (which may just be the zero-size
                 selection referring to the cursor's location) */
            while( selection && ranges.length )
            {
                var r = ranges.shift();
                if( !r.collapsed )
                {
                    // XXX: can't we set the end exactly where we want it?
                    if( fudgeRanges )
                        r.range.setEndAfter( r.range.endContainer );
                    else
                        r.range.setEnd( r.range.endContainer, r.end );
                }
                selection.addRange( r.range );
                if( r.collapsed )
                    selection.collapse( r.range.startContainer, r.start );
            }
            // And we're done
            cbSaver.restore();
            ed.endTransaction();
        }
        else {
            dump("autoRewrap: Couldn't get current editor.  Bummer.\n");
        }
    }
    catch(ex) {
        dump("autoRewrap exception: "+ex+"\n");
        ed.endTransaction();
        ed.undo(1); // undo any damage done
        gAuWrapped = false;
    }
    return;
}

/*
 * The listeners that do the magic
 * -------------------------------
 *
 * This mechanism was observed in Enigmail (http://enigmail.mozdev.org).
 * My understanding of it is a bit shaky, but hey, it seems to work.
 * The first listener gets attached onload, and then it attaches the second,
 * which gets called when the document is editable.
 */

function auwrComposeStateListener() {
}

auwrComposeStateListener.prototype = {
    NotifyComposeFieldsReady : function() {
        dump("auwrComposeStateListener.NotifyComposeFieldsReady()\n");
        try {
            gAuwrEditor = gMsgCompose.editor.QueryInterface(Components.interfaces.nsIEditor);
        }
        catch(ex){}
        if( !gAuwrEditor ) return;

        var docStateListener = new auwrDocStateListener();
        gAuwrEditor.addDocumentStateListener(docStateListener);
    }
};

function auwrDocStateListener() {
}

auwrDocStateListener.prototype = {
    QueryInterface: function (iid) {
        if (!iid.equals(Components.interfaces.nsIDocumentStateListener) &&
            !iid.equals(Components.interfaces.nsISupports))
        {
            throw Components.results.NS_ERROR_NO_INTERFACE;
        }
        return this;
    },

    NotifyDocumentStateChanged: function(nowDirty) {
        dump("auwrDocStateListener.NotifyDocumentStateChanged("+nowDirty+"): ");
        var isEmpty = gAuwrEditor.documentIsEmpty;
        var isEditable = gAuwrEditor.isDocumentEditable;
        dump( (isEmpty?"empty":"not empty")+", "
            +(isEditable?"editable":"not editable")+"\n" );

        if( isEmpty || !isEditable )
            return;

        if( !gAuWrapped )
            setTimeout(autoRewrap, 100);
    }
};

/*
 * ClipboardSaver
 * --------------
 *
 * Some of the selection manipulation done in autoRewrap() can destroy
 * the X11 selection clipboard.  This saves and restores it.
 */

function auwrClipboardSaver() {
    this.dataStack = new Array();
    this.clipBoard = Components.classes["@mozilla.org/widget/clipboard;1"]
        .getService(Components.interfaces.nsIClipboard);
    this.needed = this.clipBoard.supportsSelectionClipboard();
}

auwrClipboardSaver.prototype = {
    save: function()
    {
        if( !this.needed ) return;
        try
        {
            var transferable = Components.classes["@mozilla.org/widget/transferable;1"]
                .createInstance(Components.interfaces.nsITransferable);
            transferable.addDataFlavor("text/unicode");
            this.clipBoard.getData(transferable, this.clipBoard.kSelectionClipboard);
            var flavour = {};
            var data = {};
            var length = {};
            transferable.getAnyTransferData(flavour, data, length);
            this.dataStack.push(data);
        }
        catch(ex) { }
    },
    restore: function()
    {
        if( !this.needed ) return;
        try
        {
            var data = this.dataStack.pop();
            var pasteClipboard = Components.classes["@mozilla.org/widget/clipboardhelper;1"]
                .getService(Components.interfaces.nsIClipboardHelper);
            data = data.value.QueryInterface(Components.interfaces.nsISupportsString).data;
            pasteClipboard.copyStringToClipboard(data, this.clipBoard.kSelectionClipboard);
        }
        catch(ex) { }
    }
};

function auwrPref( prefName, prefType )
{
    var prefObj = Components.classes[ "@mozilla.org/preferences-service;1" ]
                    .getService( Components.interfaces.nsIPrefBranch );
    var prefBranch = "extensions.autorewrap.";
    try
    {
        switch( prefType )
        {
            case "int":
                return prefObj.getIntPref(prefBranch+prefName);
            case "bool":
                return prefObj.getBoolPref(prefBranch+prefName);
            case "char":
                return prefObj.getCharPref(prefBranch+prefName);
        }
    }
    catch(ex)
    {
        return null;
    }
}

/* EOF */
